__all__ = [
    'ConcurrencyLimitException',
    'ConcurrencyLimitExceededException',
]


class ConcurrencyLimitException(Exception):
    """
    Base class for all concurrency limit related exceptions.
    """
    _msg_template = None

    def __init__(self, *args, **kwargs):
        super().__init__(
            self._msg_template.format(*args, **kwargs) if self._msg_template else None,
        )


class ConcurrencyLimitExceededException(ConcurrencyLimitException):
    _msg_template = 'Exceeded the concurrency limit of {limit} executions. Waited for {timeout} seconds.'



import contextlib
import time

from ._connections import *
from .configuration import *
from .exceptions import *


__all__ = [
    'limit',
]


@contextlib.contextmanager
def limit(
        redis_configuration: RedisConfiguration,
        limit_configuration: LimitConfiguration,
):
    """
    Context manager that limits concurrency within its scope as defined on the given limit configuration. This is
    done with the logic as follows:
     - If the count of concurrently running scopes is below the configured limit, the scope is executed immediately.
     - If the count of concurrently running scopes exceeds the configured limit, the context manager wait unit it
       goes below the configured limit.
     - If the count of concurrently running scopes does not go below the configured limit with the configured
       timeout, a `ConcurrencyLimitExceededException` exception is raised.

    :param redis_configuration: Redis configuration
    :param limit_configuration: Limit configuration
    :raise ConcurrencyLimitExceededException: Configured concurrency limit not under-run within configured timeout
    :return:
    """
    client = get_redis(redis_configuration)
    start = time.time()
    executed = False

    try:
        # we loop as long as it was not possible to execute the context manager's scope.
        while not executed:
            # first we try to acquire an execution slot by increasing the concurrency counter and checking if the
            # counter is below the configured limit. if it is, we execute the scope. if it is not, we immediately
            # decrease the concurrency counter as we are not able to execute the scope.
            count = client.incr(limit_configuration.key, 1)
            client.expire(limit_configuration.key, limit_configuration.limit_expire)

            # if there is no available execution slot.
            if count > limit_configuration.limit:
                client.decr(limit_configuration.key, 1)
                elapsed = time.time() - start

                # if we are waiting longer than the configured timeout, we raise a `ConcurrencyLimitExceededException`
                # exception. executing the context manager's scope failed in this case.
                if elapsed > limit_configuration.limit_timeout:
                    raise ConcurrencyLimitExceededException(
                        limit=limit_configuration.limit,
                        timeout=limit_configuration.limit_timeout,
                    )

                # we failed to acquire an execution slot for the context manager's scope, but we want to try again.
                # however, we wait the configured interval before we do so.
                time.sleep(limit_configuration.limit_interval)

            # if we successfully acquired an execution slot for the context manager's scope.
            else:
                executed = True
                yield count

    finally:
        if executed:
            client.setnx(limit_configuration.key, 1)
            count = client.decr(limit_configuration.key, 1)

            # if there are nested concurrency contexts on the same key, and the outer concurrency timed out, we might
            # end up with a negative count. we reset to 0 if this happens.
            if count < 0:
                client.decr(limit_configuration.key, count)

        del client



import dataclasses
import redis


__all__ = [
    'RedisConfiguration',
    'LimitConfiguration',
]


@dataclasses.dataclass(eq=True, frozen=True)
class RedisConfiguration:
    """
    Redis connection configuration.
    """
    host: str = None
    "The hostname or IP of the Redis server."

    port: int = 6379
    "The Redis server port."

    db: int = 0
    "The Redis database to use."

    username: str = None
    "The authentication username."

    password: str = None
    "The authentication password."

    max_connections: int = 10
    "The maximum connections on the connection pool."

    timeout: int = 10
    "The Redis server connection timeout."

    secure: bool = False
    "Use secure connection to Redis server."

    connection_pool: redis.ConnectionPool = None
    "Use this connection pool instance instead of the other fields, if set."


@dataclasses.dataclass(eq=True, frozen=True)
class LimitConfiguration:
    """
    Concurrency limit configuration.
    """
    key: str
    "The concurrency limit group identifier."

    limit: int
    "The limit for concurrently running executions of a concurrency group."

    limit_timeout: int = 10
    "Wait time for acquiring an execution slot before raising a `ConcurrencyLimitExceededException` exception."

    limit_interval: float = 0.1
    "Wait time between attempts to acquire an execution slot."

    limit_expire: int = 60
    "Expire time for the concurrency counter on Redis."



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as fh:
    readme = fh.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='concurrency-limit',
    version=os.getenv('PACKAGE_VERSION', '0.0.0').replace('refs/tags/', ''),
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    description='A library that implements a distributed concurrency limiting mechanism using Redis as a backend.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/anexia-it/python-concurrency-limit',
    author='Andreas Stocker',
    author_email='AStocker@anexia-it.com',
    install_requires=[
        'redis>=3.5',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
    ],
)



import threading
import redis

from .configuration import *


__all__ = [
    'get_redis',
]

_connection_pool_map = {}
_connection_pool_lock = threading.Lock()


def get_redis(configuration: RedisConfiguration) -> redis.Redis:
    """
    Gets the Redis client used to store the concurrency keys.

    :param configuration: Redis connection configuration
    :return: Redis client
    """
    if configuration.connection_pool:
        return _get_redis_by_connection_pool(configuration)

    else:
        return _get_redis_by_credentials(configuration)


def _get_redis_by_credentials(configuration: RedisConfiguration) -> redis.Redis:
    """
    Gets the Redis client used to store the concurrency keys using the connection credentials on the configuration
    object. This method creates a connection pool using the credentials, als will use this connection pool for
    all subsequent calls of this method with the same configuration.

    :param configuration: Redis connection configuration
    :return: Redis client
    """
    global _connection_pool_map
    global _connection_pool_lock

    with _connection_pool_lock:
        if configuration not in _connection_pool_map:
            _connection_pool_map[configuration] = redis.BlockingConnectionPool(
                host=configuration.host,
                port=configuration.port,
                db=configuration.db,
                username=configuration.username,
                password=configuration.password,
                max_connections=configuration.max_connections,
                timeout=configuration.timeout,
                connection_class=redis.SSLConnection if configuration.secure else redis.Connection,
            )

        return redis.Redis(connection_pool=_connection_pool_map[configuration])


def _get_redis_by_connection_pool(configuration: RedisConfiguration) -> redis.Redis:
    """
    Gets the Redis client used to store the concurrency keys using the connection pool on the configuration
    object.

    :param configuration: Redis connection configuration
    :return: Redis client
    """
    return redis.Redis(connection_pool=configuration.connection_pool)
